/* Copyright 2020-2025, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "block_grid.hpp"
#include "vec.hpp"
#include <QPainter>
#include <QTimer>
#include <QPixmap>
#include <QPoint>
#include <algorithm>

namespace duck::widget {
namespace {
	const constexpr int g_icon_spacing = 3;

	[[gnu::const]]
	qtint2 scalar_to_point (int scalar, int max) {
		return {
			scalar % max,
			scalar / max
		};
	}

	[[gnu::const]]
	qtint2 to_qtint2 (const QSize& sz) {
		return {sz.width(), sz.height()};
	}
} //unnamed namespace

BlockGrid::BlockGrid (QWidget* parent, unsigned int rows, unsigned int columns, float fps) :
	QWidget(parent),
	m_block_dnd(
		this,
		{[this](QPoint pt){return locate_icon_index_at(pt);}},
		{[this](){return m_icons.size();}},
		{[this](std::size_t index)->const QPixmap&{return m_icons[index].current_frame();}}
	),
	m_icon_size(-1, -1),
	m_anim_timer(new QTimer(this)),
	m_rows(rows),
	m_columns(columns)
{
	fps = std::max(fps, 1.0f);
	m_anim_timer->start(1000.0f / fps);
	connect(m_anim_timer, &QTimer::timeout, this, &BlockGrid::advance_frame);
}

BlockGrid::~BlockGrid() noexcept = default;

QSize BlockGrid::minimumSizeHint() const {
	return sizeHint();
}

QSize BlockGrid::sizeHint() const {
	if (m_columns) {
		const auto w = icon_size().width() * m_columns + g_icon_spacing * (1 + m_columns);
		const unsigned int rows = (m_rows ? m_rows : size() / m_columns);
		const auto h = icon_size().height() * rows + g_icon_spacing * (rows + 1);

		return {static_cast<int>(w), static_cast<int>(h)};
	}
	else {
		return {-1, -1};
	}
}

void BlockGrid::set_icon_size (const QSize& sz) {
	if (sz != m_icon_size) {
		m_icon_size = sz;
		m_empty_icon.reset();
		if (m_icon_size.isValid()) {
			m_empty_icon = std::make_unique<QPixmap>(m_icon_size);
			m_empty_icon->fill(QColor(115, 115, 115, 255));
		}
	}
}

void BlockGrid::set_icon_size (int sz) {
	set_icon_size(QSize{sz, sz});
}

QSize BlockGrid::icon_size() const {
	return m_icon_size.isValid() ? m_icon_size : QSize(g_icon_spacing, g_icon_spacing);
}

void BlockGrid::push_back (AnimatedPixmap&& anim) {
	m_icons.push_back(std::move(anim));
}

std::size_t BlockGrid::size() const {
	return m_icons.size();
}

void BlockGrid::paintEvent (QPaintEvent* event) {
	QPainter painter(this);
	const int max_col = static_cast<int>(this->columns());
	int count = 0;

	for (auto& ico : m_icons) {
		auto point = g_icon_spacing + scalar_to_point(count, max_col) * (g_icon_spacing + to_qtint2(ico.size()));
		if (ico.frame_count()) {
			painter.drawPixmap(point.x(), point.y(), ico.current_frame());
		}

		++count;
	}

	const int total_squares = static_cast<int>(rows() * columns());
	for (int z = count; z < total_squares && m_empty_icon; ++z) {
		auto point = g_icon_spacing + scalar_to_point(z, max_col) * (g_icon_spacing + to_qtint2(icon_size()));
		painter.drawPixmap(point.x(), point.y(), *m_empty_icon);
	}
}

void BlockGrid::mousePressEvent(QMouseEvent* event) {
	m_block_dnd.mouse_press_event(event);
}

void BlockGrid::advance_frame() {
	for (auto& ico : m_icons) {
		ico.advance_frame();
	}
	this->update();
}

unsigned int BlockGrid::rows() const {
	if (m_rows) {
		return m_rows;
	}
	else {
		const auto widget_h = static_cast<unsigned int>(std::max(this->height(), g_icon_spacing));
		const auto ico_spacing_uns = static_cast<unsigned int>(g_icon_spacing);
		return (widget_h - ico_spacing_uns) / (icon_size().height() + ico_spacing_uns);
	}
}

unsigned int BlockGrid::columns() const {
	if (m_columns) {
		return m_columns;
	}
	else {
		const auto widget_w = static_cast<unsigned int>(std::max(this->width(), g_icon_spacing));
		const auto ico_spacing_uns = static_cast<unsigned int>(g_icon_spacing);

		return (widget_w - ico_spacing_uns) / (icon_size().width() + ico_spacing_uns);
	}
}

std::pair<int, QPoint> BlockGrid::locate_icon_index_at(QPoint pt) const {
	if (pt.x() < g_icon_spacing or pt.y() < g_icon_spacing)
		return {-1, QPoint()};

	const auto ico_spacing_uns = static_cast<unsigned int>(g_icon_spacing);
	const auto x = static_cast<unsigned int>(pt.x()) - ico_spacing_uns;
	const auto y = static_cast<unsigned int>(pt.y()) - ico_spacing_uns;
	const auto w = static_cast<unsigned int>(m_icon_size.width()) + ico_spacing_uns;
	const auto h = static_cast<unsigned int>(m_icon_size.height()) + ico_spacing_uns;

	if (x >= w * columns() or y > h * rows())
		return {-1, QPoint()};

	const unsigned int x_index = x / w;
	const unsigned int y_index = y / h;

	const auto rel_x = x - x_index * w;
	const auto rel_y = y - y_index * h;

	if (rel_x < (w - ico_spacing_uns) and rel_y < (h - ico_spacing_uns)) {
		return {
			static_cast<int>(x_index + y_index * columns()),
			QPoint(g_icon_spacing, g_icon_spacing) + pt - QPoint(static_cast<int>(rel_x), static_cast<int>(rel_y))
		};
	}
	else {
		return {-1, QPoint()};
	}
}
} //namespace duck::widget
