/* Copyright 2020-2025, Michele Santullo
 * This file is part of memoserv.
 *
 * Memoserv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Memoserv is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Memoserv.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "block_drag_and_drop.hpp"
#include <QByteArray>
#include <QDataStream>
#include <QMouseEvent>
#include <QMimeData>
#include <QDrag>
#include <QApplication>
#include <cassert>
#include <ciso646>
#include <memory>

namespace duck::widget::detail {
namespace {
constexpr char g_drag_mime[] = "application/x-dndblock";
} //unnamed namespace

BlockDragAndDrop::BlockDragAndDrop (
	QObject* parent,
	PointToIndexFunc&& func,
	std::function<std::size_t()>&& get_block_count,
	std::function<const QPixmap&(std::size_t)>&& get_pixmap
) :
	m_point_to_index(std::move(func)),
	m_get_block_count(std::move(get_block_count)),
	m_get_pixmap(std::move(get_pixmap)),
	m_parent(parent)
{
	assert(m_parent);
}

void BlockDragAndDrop::mouse_press_event (const QMouseEvent* event) {
	const auto [icon_index, block_origin]  = m_point_to_index(event->position().toPoint());
	if (icon_index < 0)
		return;
	const auto icon_index_uns = static_cast<std::size_t>(icon_index);
	if (icon_index_uns >= m_get_block_count())
		return;

	QByteArray item_data;
	QDataStream data_stream(&item_data, QIODevice::WriteOnly);
	data_stream << m_get_pixmap(icon_index_uns);

	std::unique_ptr<QMimeData> mime_data(new QMimeData);
	mime_data->setData(g_drag_mime, item_data);

	QDrag* const drag(new QDrag(m_parent));
	drag->setMimeData(mime_data.release());
	drag->setPixmap(m_get_pixmap(icon_index_uns));
	drag->setHotSpot(event->position().toPoint() - block_origin);

	switch (drag->exec(Qt::CopyAction | Qt::MoveAction, Qt::CopyAction)) {
	case Qt::MoveAction:
		break;
	case Qt::CopyAction:
		break;
	default:
		break;
	}
}
} //namespace duck::widget::detail
